@@ -36,39 +36,26 @@ except ImportError:
 
 import codecs
 import json
-import sys
 import uuid
 from copy import deepcopy
+from six import python_2_unicode_compatible, iteritems
 
 try:
     from StringIO import StringIO
 except ImportError:
     from io import StringIO
 
-from .exceptions import *
+from .exceptions import (
+    NodeIDAbsentError,
+    DuplicatedNodeIdError,
+    MultipleRootError,
+    InvalidLevelNumber,
+    LinkPastRootNodeError,
+    LoopError,
+)
 from .node import Node
 
-__author__ = 'chenxm'
-
-
-def python_2_unicode_compatible(klass):
-    """
-    (slightly modified from: http://django.readthedocs.org/en/latest/_modules/django/utils/encoding.html)
-
-    A decorator that defines __unicode__ and __str__ methods under Python 2.
-    Under Python 3 it does nothing.
-
-    To support Python 2 and 3 with a single code base, define a __str__ method
-    returning text and apply this decorator to the class.
-    """
-    if sys.version_info[0] == 2:
-        if '__str__' not in klass.__dict__:
-            raise ValueError("@python_2_unicode_compatible cannot be applied "
-                             "to %s because it doesn't define __str__()." %
-                             klass.__name__)
-        klass.__unicode__ = klass.__str__
-        klass.__str__ = lambda self: self.__unicode__().encode('utf-8')
-    return klass
+__author__ = "chenxm"
 
 
 @python_2_unicode_compatible
@@ -102,7 +89,7 @@ class Tree(object):
 
         if tree is not None:
             self.root = tree.root
-            for nid, node in tree.nodes.items():
+            for nid, node in iteritems(tree.nodes):
                 new_node = deepcopy(node) if deep else node
                 self._nodes[nid] = new_node
                 if tree.identifier != self._identifier:
@@ -130,7 +117,9 @@ class Tree(object):
         >>> subtree.tree_description
         "smart tree"
         """
-        return self.__class__(identifier=identifier, tree=self if with_tree else None, deep=deep)
+        return self.__class__(
+            identifier=identifier, tree=self if with_tree else None, deep=deep
+        )
 
     @property
     def identifier(self):
@@ -154,22 +143,28 @@ class Tree(object):
         """Return len(_nodes)"""
         return len(self._nodes)
 
-    def __setitem__(self, key, item):
-        """Set _nodes[key]"""
-        self._nodes.update({key: item})
-
     def __str__(self):
         self._reader = ""
 
         def write(line):
-            self._reader += line.decode('utf-8') + "\n"
+            self._reader += line.decode("utf-8") + "\n"
 
         self.__print_backend(func=write)
         return self._reader
 
-    def __print_backend(self, nid=None, level=ROOT, idhidden=True, filter=None,
-                        key=None, reverse=False, line_type='ascii-ex',
-                        data_property=None, func=print):
+    def __print_backend(
+        self,
+        nid=None,
+        level=ROOT,
+        idhidden=True,
+        filter=None,
+        key=None,
+        reverse=False,
+        line_type="ascii-ex",
+        data_property=None,
+        sorting=True,
+        func=print,
+    ):
         """
         Another implementation of printing tree using Stack
         Print tree structure in hierarchy style.
@@ -197,77 +192,95 @@ class Tree(object):
         # Factory for proper get_label() function
         if data_property:
             if idhidden:
+
                 def get_label(node):
                     return getattr(node.data, data_property)
+
             else:
+
                 def get_label(node):
-                    return "%s[%s]" % (getattr(node.data, data_property), node.identifier)
+                    return "%s[%s]" % (
+                        getattr(node.data, data_property),
+                        node.identifier,
+                    )
+
         else:
             if idhidden:
+
                 def get_label(node):
                     return node.tag
+
             else:
+
                 def get_label(node):
                     return "%s[%s]" % (node.tag, node.identifier)
 
         # legacy ordering
-        if key is None:
+        if sorting and key is None:
+
             def key(node):
                 return node
 
         # iter with func
-        for pre, node in self.__get(nid, level, filter, key, reverse,
-                                    line_type):
+        for pre, node in self.__get(
+            nid, level, filter, key, reverse, line_type, sorting
+        ):
             label = get_label(node)
-            func('{0}{1}'.format(pre, label).encode('utf-8'))
+            func("{0}{1}".format(pre, label).encode("utf-8"))
 
-    def __get(self, nid, level, filter_, key, reverse, line_type):
+    def __get(self, nid, level, filter_, key, reverse, line_type, sorting):
         # default filter
         if filter_ is None:
+
             def filter_(node):
                 return True
 
         # render characters
         dt = {
-            'ascii': ('|', '|-- ', '+-- '),
-            'ascii-ex': ('\u2502', '\u251c\u2500\u2500 ', '\u2514\u2500\u2500 '),
-            'ascii-exr': ('\u2502', '\u251c\u2500\u2500 ', '\u2570\u2500\u2500 '),
-            'ascii-em': ('\u2551', '\u2560\u2550\u2550 ', '\u255a\u2550\u2550 '),
-            'ascii-emv': ('\u2551', '\u255f\u2500\u2500 ', '\u2559\u2500\u2500 '),
-            'ascii-emh': ('\u2502', '\u255e\u2550\u2550 ', '\u2558\u2550\u2550 '),
+            "ascii": ("|", "|-- ", "+-- "),
+            "ascii-ex": ("\u2502", "\u251c\u2500\u2500 ", "\u2514\u2500\u2500 "),
+            "ascii-exr": ("\u2502", "\u251c\u2500\u2500 ", "\u2570\u2500\u2500 "),
+            "ascii-em": ("\u2551", "\u2560\u2550\u2550 ", "\u255a\u2550\u2550 "),
+            "ascii-emv": ("\u2551", "\u255f\u2500\u2500 ", "\u2559\u2500\u2500 "),
+            "ascii-emh": ("\u2502", "\u255e\u2550\u2550 ", "\u2558\u2550\u2550 "),
         }[line_type]
 
-        return self.__get_iter(nid, level, filter_, key, reverse, dt, [])
+        return self.__get_iter(nid, level, filter_, key, reverse, dt, [], sorting)
 
-    def __get_iter(self, nid, level, filter_, key, reverse, dt, is_last):
-        dt_vline, dt_line_box, dt_line_cor = dt
-
-        nid = self.root if (nid is None) else nid
-        if not self.contains(nid):
-            raise NodeIDAbsentError("Node '%s' is not in the tree" % nid)
+    def __get_iter(self, nid, level, filter_, key, reverse, dt, is_last, sorting):
+        dt_vertical_line, dt_line_box, dt_line_corner = dt
 
+        nid = self.root if nid is None else nid
         node = self[nid]
 
         if level == self.ROOT:
             yield "", node
         else:
-            leading = ''.join(map(lambda x: dt_vline + ' ' * 3
-            if not x else ' ' * 4, is_last[0:-1]))
-            lasting = dt_line_cor if is_last[-1] else dt_line_box
+            leading = "".join(
+                map(
+                    lambda x: dt_vertical_line + " " * 3 if not x else " " * 4,
+                    is_last[0:-1],
+                )
+            )
+            lasting = dt_line_corner if is_last[-1] else dt_line_box
             yield leading + lasting, node
 
         if filter_(node) and node.expanded:
-            children = [self[i] for i in node.successors(self._identifier) if filter_(self[i])]
+            children = [
+                self[i] for i in node.successors(self._identifier) if filter_(self[i])
+            ]
             idxlast = len(children) - 1
-            if key:
-                children.sort(key=key, reverse=reverse)
-            elif reverse:
-                children = reversed(children)
+            if sorting:
+                if key:
+                    children.sort(key=key, reverse=reverse)
+                elif reverse:
+                    children = reversed(children)
             level += 1
             for idx, child in enumerate(children):
                 is_last.append(idx == idxlast)
-                for item in self.__get_iter(child.identifier, level, filter_,
-                                            key, reverse, dt, is_last):
+                for item in self.__get_iter(
+                    child.identifier, level, filter_, key, reverse, dt, is_last, sorting
+                ):
                     yield item
                 is_last.pop()
 
@@ -289,14 +302,15 @@ class Tree(object):
         """
         if not isinstance(node, self.node_class):
             raise OSError(
-                "First parameter must be object of {}".format(self.node_class))
+                "First parameter must be object of {}".format(self.node_class)
+            )
 
         if node.identifier in self._nodes:
-            raise DuplicatedNodeIdError("Can't create node "
-                                        "with ID '%s'" % node.identifier)
+            raise DuplicatedNodeIdError(
+                "Can't create node " "with ID '%s'" % node.identifier
+            )
 
-        pid = parent.identifier if isinstance(
-            parent, self.node_class) else parent
+        pid = parent.identifier if isinstance(parent, self.node_class) else parent
 
         if pid is None:
             if self.root is not None:
@@ -304,8 +318,7 @@ class Tree(object):
             else:
                 self.root = node.identifier
         elif not self.contains(pid):
-            raise NodeIDAbsentError("Parent node '%s' "
-                                    "is not in the tree" % pid)
+            raise NodeIDAbsentError("Parent node '%s' " "is not in the tree" % pid)
 
         self._nodes.update({node.identifier: node})
         self.__update_fpointer(pid, node.identifier, self.node_class.ADD)
@@ -332,7 +345,7 @@ class Tree(object):
             raise NodeIDAbsentError("Node '%s' is not in the tree" % nid)
 
         descendant = self[nid]
-        ascendant = self[nid].bpointer
+        ascendant = self[nid]._predecessor[self._identifier]
         ascendant_level = self.level(ascendant)
 
         if level is None:
@@ -340,15 +353,18 @@ class Tree(object):
         elif nid == self.root:
             return self[nid]
         elif level >= self.level(descendant.identifier):
-            raise InvalidLevelNumber("Descendant level (level %s) must be greater \
-                                      than its ancestor's level (level %s)" % (str(self.level(descendant.identifier)), level))
+            raise InvalidLevelNumber(
+                "Descendant level (level %s) must be greater \
+                                      than its ancestor's level (level %s)"
+                % (str(self.level(descendant.identifier)), level)
+            )
 
         while ascendant is not None:
             if ascendant_level == level:
                 return self[ascendant]
             else:
                 descendant = ascendant
-                ascendant = self[descendant].bpointer
+                ascendant = self[descendant]._predecessor[self._identifier]
                 ascendant_level = self.level(ascendant)
         return None
 
@@ -398,8 +414,9 @@ class Tree(object):
             ret = self.level(nid)
         return ret
 
-    def expand_tree(self, nid=None, mode=DEPTH, filter=None, key=None,
-                    reverse=False, sorting=True):
+    def expand_tree(
+        self, nid=None, mode=DEPTH, filter=None, key=None, reverse=False, sorting=True
+    ):
         """
         Python generator to traverse the tree (or a subtree) with optional
         node filtering and sorting.
@@ -429,14 +446,21 @@ class Tree(object):
         filter = (lambda x: True) if (filter is None) else filter
         if filter(self[nid]):
             yield nid
-            queue = [self[i] for i in self[nid].successors(self._identifier) if filter(self[i])]
+            queue = [
+                self[i]
+                for i in self[nid].successors(self._identifier)
+                if filter(self[i])
+            ]
             if mode in [self.DEPTH, self.WIDTH]:
                 if sorting:
                     queue.sort(key=key, reverse=reverse)
                 while queue:
                     yield queue[0].identifier
-                    expansion = [self[i] for i in queue[0].successors(self._identifier)
-                                 if filter(self[i])]
+                    expansion = [
+                        self[i]
+                        for i in queue[0].successors(self._identifier)
+                        if filter(self[i])
+                    ]
                     if sorting:
                         expansion.sort(key=key, reverse=reverse)
                     if mode is self.DEPTH:
@@ -451,8 +475,11 @@ class Tree(object):
                 stack = stack_bw = queue
                 direction = False
                 while stack:
-                    expansion = [self[i] for i in stack[0].successors(self._identifier)
-                                 if filter(self[i])]
+                    expansion = [
+                        self[i]
+                        for i in stack[0].successors(self._identifier)
+                        if filter(self[i])
+                    ]
                     yield stack.pop(0).identifier
                     if direction:
                         expansion.reverse()
@@ -464,8 +491,7 @@ class Tree(object):
                         stack = stack_fw if direction else stack_bw
 
             else:
-                raise ValueError(
-                    "Traversal mode '{}' is not supported".format(mode))
+                raise ValueError("Traversal mode '{}' is not supported".format(mode))
 
     def filter_nodes(self, func):
         """
@@ -539,8 +565,9 @@ class Tree(object):
         if not self.contains(nid):
             raise NodeIDAbsentError("Node '%s' is not in the tree" % nid)
         if self.root == nid:
-            raise LinkPastRootNodeError("Cannot link past the root node, "
-                                        "delete it with remove_node()")
+            raise LinkPastRootNodeError(
+                "Cannot link past the root node, " "delete it with remove_node()"
+            )
         # Get the parent of the node we are linking past
         parent = self[self[nid].predecessor(self._identifier)]
         # Set the children of the node to the parent
@@ -660,9 +687,9 @@ class Tree(object):
 
         set_joint = set(new_tree._nodes) & set(self._nodes)  # joint keys
         if set_joint:
-            raise ValueError('Duplicated nodes %s exists.' % list(map(text, set_joint)))
+            raise ValueError("Duplicated nodes %s exists." % list(map(text, set_joint)))
 
-        for cid, node in new_tree.nodes.items():
+        for cid, node in iteritems(new_tree.nodes):
             if deep:
                 node = deepcopy(new_tree[node])
             self._nodes.update({cid: node})
@@ -713,8 +740,7 @@ class Tree(object):
         Return the number of removed nodes.
         """
         if not self.contains(identifier):
-            raise NodeIDAbsentError("Node '%s' "
-                                    "is not in the tree" % identifier)
+            raise NodeIDAbsentError("Node '%s' " "is not in the tree" % identifier)
 
         parent = self[identifier].predecessor(self._identifier)
 
@@ -799,24 +825,61 @@ class Tree(object):
             if filter(self[current]):
                 yield current
             # subtree() hasn't update the bpointer
-            current = self[current].predecessor(self._identifier) if self.root != current else None
+            current = (
+                self[current].predecessor(self._identifier)
+                if self.root != current
+                else None
+            )
 
-    def save2file(self, filename, nid=None, level=ROOT, idhidden=True,
-                  filter=None, key=None, reverse=False, line_type='ascii-ex', data_property=None):
+    def save2file(
+        self,
+        filename,
+        nid=None,
+        level=ROOT,
+        idhidden=True,
+        filter=None,
+        key=None,
+        reverse=False,
+        line_type="ascii-ex",
+        data_property=None,
+        sorting=True,
+    ):
         """
         Save the tree into file for offline analysis.
         """
 
         def _write_line(line, f):
-            f.write(line + b'\n')
+            f.write(line + b"\n")
 
-        handler = lambda x: _write_line(x, open(filename, 'ab'))
+        def handler(x):
+            return _write_line(x, open(filename, "ab"))
 
-        self.__print_backend(nid, level, idhidden, filter,
-                             key, reverse, line_type, data_property, func=handler)
+        self.__print_backend(
+            nid,
+            level,
+            idhidden,
+            filter,
+            key,
+            reverse,
+            line_type,
+            data_property,
+            sorting,
+            func=handler,
+        )
 
-    def show(self, nid=None, level=ROOT, idhidden=True, filter=None,
-             key=None, reverse=False, line_type='ascii-ex', data_property=None, stdout=True):
+    def show(
+        self,
+        nid=None,
+        level=ROOT,
+        idhidden=True,
+        filter=None,
+        key=None,
+        reverse=False,
+        line_type="ascii-ex",
+        data_property=None,
+        stdout=True,
+        sorting=True,
+    ):
         """
         Print the tree structure in hierarchy style.
 
@@ -825,7 +888,9 @@ class Tree(object):
         former two use the same backend to generate a string of tree structure in a
         text graph.
 
-        * Version >= 1.2.7a*: you can also specify the ``line_type`` parameter, such as 'ascii' (default), 'ascii-ex', 'ascii-exr', 'ascii-em', 'ascii-emv', 'ascii-emh') to the change graphical form.
+        * Version >= 1.2.7a*: you can also specify the ``line_type`` parameter,
+          such as 'ascii' (default), 'ascii-ex', 'ascii-exr', 'ascii-em', 'ascii-emv',
+          'ascii-emh') to the change graphical form.
 
         :param nid: the reference node to start expanding.
         :param level: the node level in the tree (root as level 0).
@@ -837,18 +902,31 @@ class Tree(object):
         :param reverse: the ``reverse`` param for sorting :class:`Node` objects in the same level.
         :param line_type:
         :param data_property: the property on the node data object to be printed.
+        :param sorting: if True perform node sorting, if False return
+            nodes in original insertion order. In latter case @key and
+            @reverse parameters are ignored.
         :return: None
         """
         self._reader = ""
 
         def write(line):
-            self._reader += line.decode('utf-8') + "\n"
+            self._reader += line.decode("utf-8") + "\n"
 
         try:
-            self.__print_backend(nid, level, idhidden, filter,
-                                 key, reverse, line_type, data_property, func=write)
+            self.__print_backend(
+                nid,
+                level,
+                idhidden,
+                filter,
+                key,
+                reverse,
+                line_type,
+                data_property,
+                sorting,
+                func=write,
+            )
         except NodeIDAbsentError:
-            print('Tree is empty')
+            print("Tree is empty")
 
         if stdout:
             print(self._reader)
@@ -865,7 +943,9 @@ class Tree(object):
 
         if nid != self.root:
             pid = self[nid].predecessor(self._identifier)
-            siblings = [self[i] for i in self[pid].successors(self._identifier) if i != nid]
+            siblings = [
+                self[i] for i in self[pid].successors(self._identifier) if i != nid
+            ]
 
         return siblings
 
@@ -885,10 +965,17 @@ class Tree(object):
         else:
             try:
                 level = int(level)
-                return len([node for node in self.all_nodes_itr() if self.level(node.identifier) == level])
-            except:
+                return len(
+                    [
+                        node
+                        for node in self.all_nodes_itr()
+                        if self.level(node.identifier) == level
+                    ]
+                )
+            except Exception:
                 raise TypeError(
-                    "level should be an integer instead of '%s'" % type(level))
+                    "level should be an integer instead of '%s'" % type(level)
+                )
 
     def subtree(self, nid, identifier=None):
         """
@@ -930,8 +1017,8 @@ class Tree(object):
         :return: None
         """
         cn = self[nid]
-        for attr, val in attrs.items():
-            if attr == 'identifier':
+        for attr, val in iteritems(attrs):
+            if attr == "identifier":
                 # Updating node id meets following contraints:
                 # * Update node identifier property
                 # * Update parent's followers
@@ -939,13 +1026,15 @@ class Tree(object):
                 # * Update tree registration of var _nodes
                 # * Update tree root if necessary
                 cn = self._nodes.pop(nid)
-                setattr(cn, 'identifier', val)
+                setattr(cn, "identifier", val)
                 self._nodes[val] = cn
 
                 if cn.predecessor(self._identifier) is not None:
                     self[cn.predecessor(self._identifier)].update_successors(
-                        nid, mode=self.node_class.REPLACE, replace=val,
-                        tree_id=self._identifier
+                        nid,
+                        mode=self.node_class.REPLACE,
+                        replace=val,
+                        tree_id=self._identifier,
                     )
 
                 for fp in cn.successors(self._identifier):
@@ -973,49 +1062,67 @@ class Tree(object):
 
             for elem in queue:
                 tree_dict[ntag]["children"].append(
-                    self.to_dict(elem.identifier, with_data=with_data, sort=sort, reverse=reverse))
+                    self.to_dict(
+                        elem.identifier, with_data=with_data, sort=sort, reverse=reverse
+                    )
+                )
             if len(tree_dict[ntag]["children"]) == 0:
-                tree_dict = self[nid].tag if not with_data else \
-                    {ntag: {"data": self[nid].data}}
+                tree_dict = (
+                    self[nid].tag if not with_data else {ntag: {"data": self[nid].data}}
+                )
             return tree_dict
 
     def to_json(self, with_data=False, sort=True, reverse=False):
         """To format the tree in JSON format."""
         return json.dumps(self.to_dict(with_data=with_data, sort=sort, reverse=reverse))
 
-    def to_graphviz(self, filename=None, shape='circle', graph='digraph'):
+    def to_graphviz(
+        self,
+        filename=None,
+        shape="circle",
+        graph="digraph",
+        filter=None,
+        key=None,
+        reverse=False,
+        sorting=True,
+    ):
         """Exports the tree in the dot format of the graphviz software"""
         nodes, connections = [], []
         if self.nodes:
-
-            for n in self.expand_tree(mode=self.WIDTH):
+            for n in self.expand_tree(
+                mode=self.WIDTH,
+                filter=filter,
+                key=key,
+                reverse=reverse,
+                sorting=sorting,
+            ):
                 nid = self[n].identifier
-                state = '"{0}" [label="{1}", shape={2}]'.format(
-                    nid, self[n].tag, shape)
+                state = '"{0}" [label="{1}", shape={2}]'.format(nid, self[n].tag, shape)
                 nodes.append(state)
 
                 for c in self.children(nid):
                     cid = c.identifier
-                    connections.append('"{0}" -> "{1}"'.format(nid, cid))
+                    edge = "->" if graph == "digraph" else "--"
+                    connections.append(('"{0}" ' + edge + ' "{1}"').format(nid, cid))
 
         # write nodes and connections to dot format
         is_plain_file = filename is not None
         if is_plain_file:
-            f = codecs.open(filename, 'w', 'utf-8')
+            f = codecs.open(filename, "w", "utf-8")
         else:
             f = StringIO()
 
-        f.write(graph + ' tree {\n')
+        f.write(graph + " tree {\n")
         for n in nodes:
-            f.write('\t' + n + '\n')
+            f.write("\t" + n + "\n")
 
         if len(connections) > 0:
-            f.write('\n')
+            f.write("\n")
 
         for c in connections:
-            f.write('\t' + c + '\n')
+            f.write("\t" + c + "\n")
 
-        f.write('}')
+        f.write("}")
 
         if not is_plain_file:
             print(f.getvalue())
